import 'dart:math';

import '../entities/cell.dart';
import '../utils/random_pick_grid.dart';
import '../provider/data.dart';

class BoardUtils {

  static printGrid(List cells) {
    print('');
    print('-------');
    for (var rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String row = '';
      for (var colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row += cells[rowIndex][colIndex].value.toString();
      }
      print(row);
    }
    print('-------');
    print('');
  }


  static Future<void> pickGrid(Data myProvider) async {
    int size = myProvider.size;

    String grid;
    RandomPickGrid randomPickGrid;

    randomPickGrid = RandomPickGrid();
    await randomPickGrid.init(myProvider.level, size);

    if (randomPickGrid.grid != null) {
      grid = randomPickGrid.grid;
    }

    if (grid.length == pow(size, 4)) {
      myProvider.updateCells = BoardUtils.createBoardFromTemplate(grid);
    }
  }


  static List createEmptyBoard(int size) {
    int index = 0;
    List cells = [];
    for (var rowIndex = 0; rowIndex < pow(size, 2); rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < pow(size, 2); colIndex++) {
        row.add(Cell(0, false));
      }
      cells.add(row);
    }

    return cells;
  }


  static List createBoardFromTemplate(String grid) {
    List cells = [];
    int size = int.parse(pow(grid.length, 1/4).toStringAsFixed(0));
    int sideLength = pow(size, 2);

    int index = 0;
    for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < sideLength; colIndex++) {
        int value = int.parse(grid[index++]);
        row.add(Cell(value, (value != 0)));
      }
      cells.add(row);
    }

    List<String> allowedFlip = ['', 'horizontal', 'vertical'];
    List<String> allowedRotate = ['', 'left', 'right'];

    var rand = new Random();
    String flip = allowedFlip[rand.nextInt(allowedFlip.length)];
    String rotate = allowedRotate[rand.nextInt(allowedRotate.length)];

    switch(flip) {
      case 'horizontal': {
        List transformedBoard = createEmptyBoard(size);
        for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
          for (var colIndex = 0; colIndex < sideLength; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[sideLength - rowIndex - 1][colIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'vertical': {
        List transformedBoard = createEmptyBoard(size);
        for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
          for (var colIndex = 0; colIndex < sideLength; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[rowIndex][sideLength - colIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    switch(rotate) {
      case 'left': {
        List transformedBoard = createEmptyBoard(size);
        for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
          for (var colIndex = 0; colIndex < sideLength; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[colIndex][sideLength - rowIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'right': {
        List transformedBoard = createEmptyBoard(size);
        for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
          for (var colIndex = 0; colIndex < sideLength; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[sideLength - colIndex - 1][rowIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    // Fix cells fixed states
    for (var rowIndex = 0; rowIndex < sideLength; rowIndex++) {
      for (var colIndex = 0; colIndex < sideLength; colIndex++) {
        cells[rowIndex][colIndex].isFixed = (cells[rowIndex][colIndex].value != 0) ? true : false;
      }
    }

    printGrid(cells);

    return cells;
  }


  static bool checkBoardIsSolved(Data myProvider) {
    List cells = myProvider.cells;
    int size = myProvider.size;
    int sideLength = pow(size, 2);

    bool isSolved = true;

    // reset conflict states
    for (var row = 0; row < sideLength; row++) {
      for (var col = 0; col < sideLength; col++) {
        cells[row][col].conflictsCount = 0;
      }
    }

    // check grid is fully completed
    for (var row = 0; row < sideLength; row++) {
      for (var col = 0; col < sideLength; col++) {
        if (cells[row][col].value == 0) {
          isSolved = false;
        }
      }
    }

    // check lines does not contains a value twice
    for (var row = 0; row < sideLength; row++) {
      List values = [];
      for (var col = 0; col < sideLength; col++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('line ' + row.toString() + ' contains duplicates');
        // Add line to cells in conflict
        for (var col = 0; col < sideLength; col++) {
          cells[row][col].conflictsCount++;
        }
        isSolved = false;
      }
    }

    // check columns does not contains a value twice
    for (var col = 0; col < sideLength; col++) {
      List values = [];
      for (var row = 0; row < sideLength; row++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('column ' + col.toString() + ' contains duplicates');
        // Add column to cells in conflict
        for (var row = 0; row < sideLength; row++) {
          cells[row][col].conflictsCount++;
        }
        isSolved = false;
      }
    }

    // check blocks does not contains a value twice
    for (var blockRow = 0; blockRow < size; blockRow++) {
      for (var blockCol = 0; blockCol < size; blockCol++) {
        List values = [];

        for (var rowInBlock = 0; rowInBlock < size; rowInBlock++) {
          for (var colInBlock = 0; colInBlock < size; colInBlock++) {
            int row = (blockRow * size) + rowInBlock;
            int col = (blockCol * size) + colInBlock;
            int value = cells[row][col].value;
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          print('block [' + blockCol.toString() + ',' + blockRow.toString() + '] contains duplicates');
          // Add blocks to cells in conflict
          for (var rowInBlock = 0; rowInBlock < size; rowInBlock++) {
            for (var colInBlock = 0; colInBlock < size; colInBlock++) {
              int row = (blockRow * size) + rowInBlock;
              int col = (blockCol * size) + colInBlock;
              cells[row][col].conflictsCount++;
            }
          }
          isSolved = false;
        }
      }
    }

    if (isSolved) {
      print('-> ok sudoku solved!');
    }

    return isSolved;
  }

}
